/*@ ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
 **
 **    Copyright (c) 2007 - 2008, 
 **        Dave A. May [dave.mayhem23@gmail.com]
 **        School of Mathematical Sciences, 
 **        Monash University
 **        Victoria, 3800
 **        Australia
 **
 **    Copyright (c) 2009 - 2010, 
 **        Dave A. May [dave.mayhem23@gmail.com]
 **        Geophysical Fluid Dynamics, 
 **        Department of Earth Sciences,
 **        ETH Zürich,
 **        Switzerland
 **
 **    Project:       PetscExt_v3.1
 **    Description:   A collection of extensions to the PETSc library.
 **
 **    All rights reserved.
 **    Redistribution and use in source and binary forms, with or without modification,
 **    are permitted provided that the following conditions are met:
 **
 **        * Redistributions of source code must retain the above copyright notice,
 **              this list of conditions and the following disclaimer.
 **        * Redistributions in binary form must reproduce the above copyright
 **              notice, this list of conditions and the following disclaimer in the
 **              documentation and/or other materials provided with the distribution.
 **        * Neither the name of the Monash University nor the names of its contributors
 **              may be used to endorse or promote products derived from this software
 **              without specific prior written permission.
 **
 **    THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 **    "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO,
 **    THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
 **    PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS
 **    BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 **    CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 **    SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 **    HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 **    LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT 
 **    OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 **
 **
 **    $Id$
 **
 ** ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~@*/


#include <petsc.h>
#include <petscvec.h>
#include <petscmat.h>
#include <private/matimpl.h>

#include "private/vec/petscvec-block.h"
#include "private/mat/petscmat-block.h"

#include "mat-schur-impl.h"
#include "private/mat/mat-schur.h"
#include "mat-schur-ops.h"


extern const char *_MatSchurComplementType[];

/* for logging */
PetscLogStage __PETScExt_MatSchur_MatMult_Stage;
PetscLogStage __PETScExt_MatSchur_MatMultTranspose_Stage;

PetscLogEvent __PETScExt_MatSchur_MatMults_Event;
PetscLogEvent __PETScExt_MatSchur_A11_inverse_Event;
PetscLogEvent __PETScExt_MatSchur_A22_inverse_Event;


/* 0*/
PetscErrorCode MatSetValues_Schur(Mat mat,PetscInt m,const PetscInt idxm[],PetscInt n,const PetscInt idxn[],const PetscScalar v[],InsertMode addv)
{
	PetscFunctionBegin; SETERRQ1( PETSC_ERR_SUP,"Operation: %s is not implemented", __func__ );

	PetscFunctionReturn(0);
}

PetscErrorCode MatGetRow_Schur(Mat mat,PetscInt row,PetscInt *ncols,PetscInt *cols[],PetscScalar *vals[])
{
	PetscFunctionBegin; SETERRQ1( PETSC_ERR_SUP,"Operation: %s is not implemented", __func__ );
	
	PetscFunctionReturn(0);
}

PetscErrorCode MatRestoreRow_Schur(Mat mat,PetscInt row,PetscInt *ncols,PetscInt *cols[],PetscScalar *vals[])
{
	PetscFunctionBegin; SETERRQ1( PETSC_ERR_SUP,"Operation: %s is not implemented", __func__ );

	PetscFunctionReturn(0);
}


//MatMult_Schur
/*
"MatSchur_A11": y = S_11 x
Solve A11 {t1} = {x}
{y} = A21 {t1}
{y} <- -{y}
{y} = {y} + A22 {x}
{y} <- -{y}
*/
PetscErrorCode MatMult_Schur_11(Mat mat,Vec x,Vec y)
{
	Mat_Schur s = (Mat_Schur)mat->data;
	PetscScalar final_scale;
	PetscTruth apply;
	
	PetscFunctionBegin; 
	
	PetscLogStagePush( __PETScExt_MatSchur_MatMult_Stage );
	
	PetscLogEventBegin(__PETScExt_MatSchur_MatMults_Event,0,0,0,0);
	/* this will save us doing one extra vector scale */
	apply = PETSC_FALSE;
	if (s->A22!=PETSC_NULL && s->scale_set == PETSC_TRUE ) {
		final_scale = - s->alpha;
		apply = PETSC_TRUE;
	}
	else if (s->A22!=PETSC_NULL && s->scale_set == PETSC_FALSE ) {
		final_scale = -1.0;
		apply = PETSC_TRUE;
	}
	else if (s->A22==PETSC_NULL && s->scale_set==PETSC_TRUE ) {
		final_scale = s->alpha;
		apply = PETSC_TRUE;
	}
	
	
	MatMult( s->A12, x, s->t1a );
	PetscLogEventEnd(__PETScExt_MatSchur_MatMults_Event,0,0,0,0);
	
	PetscLogEventBegin(__PETScExt_MatSchur_A11_inverse_Event,0,0,0,0);
	KSPSolve( s->ksp, s->t1a, s->t1 );
	PetscLogEventEnd(__PETScExt_MatSchur_A11_inverse_Event,0,0,0,0);
	
	PetscLogEventBegin(__PETScExt_MatSchur_MatMults_Event,0,0,0,0);
	MatMult( s->A21, s->t1, y );
	
	if( s->A22 != PETSC_NULL ) {
		VecScale( y, -1.0 );
		
		MatMultAdd( s->A22,x, y,y );
	}
	
	if (apply) {
		VecScale( y, final_scale );
	}
	PetscLogEventEnd(__PETScExt_MatSchur_MatMults_Event,0,0,0,0);
	
	PetscLogStagePop();
	
	PetscFunctionReturn(0);
}

/*
"MatSchur_A22": y = S_22 x
Solve A22 {t2} = {x}
{y} = A12 {t2}

{y} <- -{y}
{y} = {y} + A11 {x}
{y} <- -{y}
*/
PetscErrorCode MatMult_Schur_22(Mat mat,Vec x,Vec y)
{
	Mat_Schur s = (Mat_Schur)mat->data;
	PetscScalar final_scale;
	PetscTruth apply;
	
	PetscFunctionBegin; 
	
	PetscLogStagePush( __PETScExt_MatSchur_MatMult_Stage );
	
	PetscLogEventBegin(__PETScExt_MatSchur_MatMults_Event,0,0,0,0);
	/* this will save us doing one extra vector scale */
	apply = PETSC_FALSE;
	if (s->A11!=PETSC_NULL && s->scale_set == PETSC_TRUE ) {
		final_scale = - s->alpha;
		apply = PETSC_TRUE;
	}
	else if (s->A11!=PETSC_NULL && s->scale_set == PETSC_FALSE ) {
		final_scale = -1.0;
		apply = PETSC_TRUE;
	}
	else if (s->A11==PETSC_NULL && s->scale_set==PETSC_TRUE ) {
		final_scale = s->alpha;
		apply = PETSC_TRUE;
	}
	
	
	MatMult( s->A21, x, s->t2a );
	PetscLogEventEnd(__PETScExt_MatSchur_MatMults_Event,0,0,0,0);
	
	PetscLogEventBegin(__PETScExt_MatSchur_A22_inverse_Event,0,0,0,0);
	KSPSolve( s->ksp, s->t2a, s->t2 );
	PetscLogEventEnd(__PETScExt_MatSchur_A22_inverse_Event,0,0,0,0);
	
	PetscLogEventBegin(__PETScExt_MatSchur_MatMults_Event,0,0,0,0);
	MatMult( s->A12, s->t2, y );
	
	if( s->A11 != PETSC_NULL ) {
		VecScale( y, -1.0 );
		
		MatMultAdd( s->A11,x, y,y );
	}
	
	
	if (apply) {
		VecScale( y, final_scale );
	}
	
	PetscLogEventEnd(__PETScExt_MatSchur_MatMults_Event,0,0,0,0);
	
	PetscLogStagePop();
	
	PetscFunctionReturn(0);
}



PetscErrorCode MatMultAdd_Schur(Mat mat,Vec v1,Vec v2,Vec v3)
{
	PetscFunctionBegin;
	
	MatMult( mat,v1, v2 );
	VecAXPY( v3, 1.0, v2 );
	
	PetscFunctionReturn(0);
}


/* 5*/
PetscErrorCode MatMultTranspose_Schur_11(Mat mat,Vec x,Vec y)
{
	Mat_Schur s = (Mat_Schur)mat->data;
	PetscScalar final_scale;
	PetscTruth apply;
	
	PetscFunctionBegin; 
	
	PetscLogStagePush( __PETScExt_MatSchur_MatMultTranspose_Stage );
	
	PetscLogEventBegin(__PETScExt_MatSchur_MatMults_Event,0,0,0,0);
	
	/* this will save us doing one extra vector scale */
	apply = PETSC_FALSE;
	if (s->A22!=PETSC_NULL && s->scale_set == PETSC_TRUE ) {
		final_scale = - s->alpha;
		apply = PETSC_TRUE;
	}
	else if (s->A22!=PETSC_NULL && s->scale_set == PETSC_FALSE ) {
		final_scale = -1.0;
		apply = PETSC_TRUE;
	}
	else if (s->A22==PETSC_NULL && s->scale_set==PETSC_TRUE ) {
		final_scale = s->alpha;
		apply = PETSC_TRUE;
	}
	
	
	
	MatMultTranspose( s->A21, x, s->t1a );
	PetscLogEventEnd(__PETScExt_MatSchur_MatMults_Event,0,0,0,0);
	
	PetscLogEventBegin(__PETScExt_MatSchur_A11_inverse_Event,0,0,0,0);
	KSPSolveTranspose( s->ksp, s->t1a, s->t1 );
	PetscLogEventEnd(__PETScExt_MatSchur_A11_inverse_Event,0,0,0,0);
	
	PetscLogEventBegin(__PETScExt_MatSchur_MatMults_Event,0,0,0,0);
	MatMultTranspose( s->A12, s->t1, y );
	
	if( s->A22 != PETSC_NULL ) {
		VecScale( y, -1.0 );
		
		MatMultTransposeAdd( s->A22,x, y,y );
		
	}
	
	
	if (apply) {
		VecScale( y, final_scale );
	}
	PetscLogEventEnd(__PETScExt_MatSchur_MatMults_Event,0,0,0,0);
	
	PetscLogStagePop();
	
	PetscFunctionReturn(0);
}

PetscErrorCode MatMultTranspose_Schur_22(Mat mat,Vec x,Vec y)
{
	Mat_Schur s = (Mat_Schur)mat->data;
	PetscScalar final_scale;
	PetscTruth apply;
	
	PetscFunctionBegin; 
	
	PetscLogStagePush( __PETScExt_MatSchur_MatMultTranspose_Stage );
	
	PetscLogEventBegin(__PETScExt_MatSchur_MatMults_Event,0,0,0,0);
	
	/* this will save us doing one extra vector scale */
	apply = PETSC_FALSE;
	if (s->A11!=PETSC_NULL && s->scale_set == PETSC_TRUE ) {
		final_scale = - s->alpha;
		apply = PETSC_TRUE;
	}
	else if (s->A11!=PETSC_NULL && s->scale_set == PETSC_FALSE ) {
		final_scale = -1.0;
		apply = PETSC_TRUE;
	}
	else if (s->A11==PETSC_NULL && s->scale_set==PETSC_TRUE ) {
		final_scale = s->alpha;
		apply = PETSC_TRUE;
	}
	
	MatMultTranspose( s->A12, x, s->t2a );
	PetscLogEventEnd(__PETScExt_MatSchur_MatMults_Event,0,0,0,0);
	
	PetscLogEventBegin(__PETScExt_MatSchur_A22_inverse_Event,0,0,0,0);
	KSPSolveTranspose( s->ksp, s->t2a, s->t2 );
	PetscLogEventEnd(__PETScExt_MatSchur_A22_inverse_Event,0,0,0,0);
	
	PetscLogEventBegin(__PETScExt_MatSchur_MatMults_Event,0,0,0,0);
	MatMultTranspose( s->A21, s->t2, y );
	
	if( s->A11 != PETSC_NULL ) {
		VecScale( y, -1.0 );
		
		MatMultTransposeAdd( s->A11,x, y,y );
		
	}
	
	if (apply) {
		VecScale( y, final_scale );
	}
	PetscLogEventEnd(__PETScExt_MatSchur_MatMults_Event,0,0,0,0);
	
	PetscLogStagePop();
	
	PetscFunctionReturn(0);
}





PetscErrorCode MatMultTransposeAdd_Schur(Mat mat,Vec v1,Vec v2,Vec v3)
{
	PetscFunctionBegin; 
	
	MatMultTranspose( mat, v1, v2 );
	VecAXPY( v3, 1.0, v2 );
	
	PetscFunctionReturn(0);
}
PetscErrorCode MatSolve_Schur(Mat mat,Vec b,Vec x)
{
	PetscFunctionBegin; SETERRQ1( PETSC_ERR_SUP,"Operation: %s is not implemented", __func__ );
	
	PetscFunctionReturn(0);
}
PetscErrorCode MatSolveAdd_Schur(Mat mat,Vec b,Vec y,Vec x)
{
	PetscFunctionBegin; SETERRQ1( PETSC_ERR_SUP,"Operation: %s is not implemented", __func__ );
	
	PetscFunctionReturn(0);
}

PetscErrorCode MatSolveTranspose_Schur(Mat mat,Vec b,Vec x)
{
	PetscFunctionBegin; SETERRQ1( PETSC_ERR_SUP,"Operation: %s is not implemented", __func__ );
	
	PetscFunctionReturn(0);
}

/*10*/
PetscErrorCode MatSolveTransposeAdd_Schur(Mat mat,Vec b,Vec y,Vec x)
{
	PetscFunctionBegin; SETERRQ1( PETSC_ERR_SUP,"Operation: %s is not implemented", __func__ );
	
	PetscFunctionReturn(0);
}
PetscErrorCode MatLUFactor_Schur(Mat mat,IS row,IS col,MatFactorInfo *info)
{
	PetscFunctionBegin; SETERRQ1( PETSC_ERR_SUP,"Operation: %s is not implemented", __func__ );
	
	PetscFunctionReturn(0);
}
PetscErrorCode MatCholeskyFactor_Schur(Mat mat,IS perm,MatFactorInfo *info)
{
	PetscFunctionBegin; SETERRQ1( PETSC_ERR_SUP,"Operation: %s is not implemented", __func__ );
	
	PetscFunctionReturn(0);
}
PetscErrorCode MatRelax_Schur(Mat mat,Vec b,PetscReal omega,MatSORType flag,PetscReal shift,PetscInt its,PetscInt lits,Vec x)
{
	PetscFunctionBegin; SETERRQ1( PETSC_ERR_SUP,"Operation: %s is not implemented", __func__ );
	
	PetscFunctionReturn(0);
}
PetscErrorCode MatTranspose_Schur(Mat mat,MatReuse reuse,Mat *B)
{
	PetscFunctionBegin; SETERRQ1( PETSC_ERR_SUP,"Operation: %s is not implemented", __func__ );
	
	PetscFunctionReturn(0);
}

/*15*/

PetscErrorCode MatGetInfo_Schur(Mat mat,MatInfoType flag,MatInfo *info)
{
	PetscFunctionBegin; SETERRQ1( PETSC_ERR_SUP,"Operation: %s is not implemented", __func__ );
	
	PetscFunctionReturn(0);
}
PetscErrorCode MatEqual_Schur(Mat A,Mat B,PetscTruth *flg)
{
	PetscFunctionBegin; SETERRQ1( PETSC_ERR_SUP,"Operation: %s is not implemented", __func__ );
	
	PetscFunctionReturn(0);
}
PetscErrorCode MatGetDiagonal_Schur(Mat mat,Vec v)
{
	PetscFunctionBegin; SETERRQ1( PETSC_ERR_SUP,"Operation: %s is not implemented", __func__ );
	
	PetscFunctionReturn(0);
}
PetscErrorCode MatDiagonalScale_Schur(Mat mat,Vec l,Vec r)
{
	PetscFunctionBegin; SETERRQ1( PETSC_ERR_SUP,"Operation: %s is not implemented", __func__ );
	
	PetscFunctionReturn(0);
}
PetscErrorCode MatNorm_Schur(Mat mat,NormType type,PetscReal *nrm)
{
	PetscFunctionBegin; SETERRQ1( PETSC_ERR_SUP,"Operation: %s is not implemented", __func__ );
	
	PetscFunctionReturn(0);
}

/*20*/
PetscErrorCode MatAssemblyBegin_Schur(Mat mat,MatAssemblyType type)
{
	Mat_Schur s = (Mat_Schur)mat->data;
	PetscTruth assembled;
	
	PetscFunctionBegin; 
	
	
	if( s->A11 != PETSC_NULL ) {
		MatAssembled( s->A11, &assembled );
		if( assembled == PETSC_FALSE ) {	MatAssemblyBegin( s->A11, MAT_FINAL_ASSEMBLY ); 	}
	}
	if( s->A12 != PETSC_NULL ) {
		MatAssembled( s->A12, &assembled );
		if( assembled == PETSC_FALSE ) {	MatAssemblyBegin( s->A12, MAT_FINAL_ASSEMBLY ); 	}
	}
	if( s->A21 != PETSC_NULL ) {
		MatAssembled( s->A21, &assembled );
		if( assembled == PETSC_FALSE ) {	MatAssemblyBegin( s->A21, MAT_FINAL_ASSEMBLY ); 	}
	}
	if( s->A22 != PETSC_NULL ) {
		MatAssembled( s->A22, &assembled );
		if( assembled == PETSC_FALSE ) {	MatAssemblyBegin( s->A22, MAT_FINAL_ASSEMBLY ); 	}
	}
	
	PetscFunctionReturn(0);
}
PetscErrorCode MatAssemblyEnd_Schur(Mat mat,MatAssemblyType type)
{
	Mat_Schur s = (Mat_Schur)mat->data;
	PetscTruth assembled;
	
	PetscFunctionBegin; 
	
	
	if( s->A11 != PETSC_NULL ) {
		MatAssembled( s->A11, &assembled );
		if( assembled == PETSC_FALSE ) {	MatAssemblyEnd( s->A11, MAT_FINAL_ASSEMBLY ); 	}
	}
	if( s->A12 != PETSC_NULL ) {
		MatAssembled( s->A12, &assembled );
		if( assembled == PETSC_FALSE ) {	MatAssemblyEnd( s->A12, MAT_FINAL_ASSEMBLY ); 	}
	}
	if( s->A21 != PETSC_NULL ) {
		MatAssembled( s->A21, &assembled );
		if( assembled == PETSC_FALSE ) {	MatAssemblyEnd( s->A21, MAT_FINAL_ASSEMBLY ); 	}
	}
	if( s->A22 != PETSC_NULL ) {
		MatAssembled( s->A22, &assembled );
		if( assembled == PETSC_FALSE ) {	MatAssemblyEnd( s->A22, MAT_FINAL_ASSEMBLY ); 	}
	}
	
	PetscFunctionReturn(0);
}
PetscErrorCode MatCompress_Schur(Mat mat)
{
	PetscFunctionBegin; SETERRQ1( PETSC_ERR_SUP,"Operation: %s is not implemented", __func__ );
	
	PetscFunctionReturn(0);
}
PetscErrorCode MatSetOption_Schur(Mat mat,MatOption op)
{
	PetscFunctionBegin; SETERRQ1( PETSC_ERR_SUP,"Operation: %s is not implemented", __func__ );
	
	PetscFunctionReturn(0);
}
PetscErrorCode MatZeroEntries_Schur(Mat mat)
{
	PetscFunctionBegin; SETERRQ1( PETSC_ERR_SUP,"Operation: %s is not implemented", __func__ );
	
	PetscFunctionReturn(0);
}

/*25*/
PetscErrorCode MatZeroRows_Schur(Mat mat,PetscInt numRows,const PetscInt rows[],PetscScalar diag)
{
	PetscFunctionBegin; SETERRQ1( PETSC_ERR_SUP,"Operation: %s is not implemented", __func__ );
	
	PetscFunctionReturn(0);
}
PetscErrorCode MatLUFactorSymbolic_Schur(Mat mat,IS row,IS col,MatFactorInfo *info,Mat *fact)
{
	PetscFunctionBegin; SETERRQ1( PETSC_ERR_SUP,"Operation: %s is not implemented", __func__ );
	
	PetscFunctionReturn(0);
}
PetscErrorCode MatLUFactorNumeric_Schur(Mat mat,MatFactorInfo *info,Mat *fact)
{
	PetscFunctionBegin; SETERRQ1( PETSC_ERR_SUP,"Operation: %s is not implemented", __func__ );
	
	PetscFunctionReturn(0);
}
PetscErrorCode MatCholeskyFactorSymbolic_Schur(Mat mat,IS perm,MatFactorInfo *info,Mat *fact)
{
	PetscFunctionBegin; SETERRQ1( PETSC_ERR_SUP,"Operation: %s is not implemented", __func__ );
	
	PetscFunctionReturn(0);
}
PetscErrorCode MatCholeskyFactorNumeric_Schur(Mat mat,MatFactorInfo *info,Mat *fact)
{
	PetscFunctionBegin; SETERRQ1( PETSC_ERR_SUP,"Operation: %s is not implemented", __func__ );
	
	PetscFunctionReturn(0);
}

/*30*/
PetscErrorCode MatSetUpPreallocation_Schur(Mat B)
{
	PetscFunctionBegin; SETERRQ1( PETSC_ERR_SUP,"Operation: %s is not implemented", __func__ );
	
	PetscFunctionReturn(0);
}
PetscErrorCode MatILUFactorSymbolic_Schur(Mat mat,IS row,IS col,MatFactorInfo *info,Mat *fact)
{
	PetscFunctionBegin; SETERRQ1( PETSC_ERR_SUP,"Operation: %s is not implemented", __func__ );
	
	PetscFunctionReturn(0);
}
PetscErrorCode MatICCFactorSymbolic_Schur(Mat mat,IS perm,MatFactorInfo *info,Mat *fact)
{
	PetscFunctionBegin; SETERRQ1( PETSC_ERR_SUP,"Operation: %s is not implemented", __func__ );
	
	PetscFunctionReturn(0);
}
PetscErrorCode MatGetArray_Schur(Mat mat,PetscScalar *v[])
{
	PetscFunctionBegin; SETERRQ1( PETSC_ERR_SUP,"Operation: %s is not implemented", __func__ );
	
	PetscFunctionReturn(0);
}
PetscErrorCode MatRestoreArray_Schur(Mat mat,PetscScalar *v[])
{
	PetscFunctionBegin; SETERRQ1( PETSC_ERR_SUP,"Operation: %s is not implemented", __func__ );
	
	PetscFunctionReturn(0);
}

/*35*/
PetscErrorCode MatDuplicate_Schur(Mat mat,MatDuplicateOption op,Mat *M)
{
	
	PetscFunctionBegin; 
	
	PetscFunctionReturn(0);
}
PetscErrorCode MatForwardSolve_Schur(Mat mat,Vec b,Vec x)
{
	PetscFunctionBegin; SETERRQ1( PETSC_ERR_SUP,"Operation: %s is not implemented", __func__ );
	
	PetscFunctionReturn(0);
}
PetscErrorCode MatBackwardSolve_Schur(Mat mat,Vec b,Vec x)
{
	PetscFunctionBegin; SETERRQ1( PETSC_ERR_SUP,"Operation: %s is not implemented", __func__ );
	
	PetscFunctionReturn(0);
}
PetscErrorCode MatILUFactor_Schur(Mat mat,IS row,IS col,MatFactorInfo *info)
{
	PetscFunctionBegin; SETERRQ1( PETSC_ERR_SUP,"Operation: %s is not implemented", __func__ );
	
	PetscFunctionReturn(0);
}
PetscErrorCode MatICCFactor_Schur(Mat mat,IS row,MatFactorInfo* info)
{
	PetscFunctionBegin; SETERRQ1( PETSC_ERR_SUP,"Operation: %s is not implemented", __func__ );
	
	PetscFunctionReturn(0);
}

/*40*/
PetscErrorCode MatAXPY_Schur(Mat Y,PetscScalar a,Mat X,MatStructure str)
{
	PetscFunctionBegin; SETERRQ1( PETSC_ERR_SUP,"Operation: %s is not implemented", __func__ );
	
	PetscFunctionReturn(0);
}
PetscErrorCode MatGetSubMatrices_Schur(Mat mat,PetscInt n,const IS irow[],const IS icol[],MatReuse scall,Mat *submat[])
{
	PetscFunctionBegin; SETERRQ1( PETSC_ERR_SUP,"Operation: %s is not implemented", __func__ );
	
	PetscFunctionReturn(0);
}
PetscErrorCode MatIncreaseOverlap_Schur(Mat mat,PetscInt n,IS is[],PetscInt ov)
{
	PetscFunctionBegin; SETERRQ1( PETSC_ERR_SUP,"Operation: %s is not implemented", __func__ );
	
	PetscFunctionReturn(0);
}
PetscErrorCode MatGetValues_Schur(Mat mat,PetscInt m,const PetscInt idxm[],PetscInt n,const PetscInt idxn[],PetscScalar v[])
{
	PetscFunctionBegin; SETERRQ1( PETSC_ERR_SUP,"Operation: %s is not implemented", __func__ );
	
	PetscFunctionReturn(0);
}
PetscErrorCode MatCopy_Schur(Mat A,Mat B,MatStructure str)
{
	PetscFunctionBegin; SETERRQ1( PETSC_ERR_SUP,"Operation: %s is not implemented", __func__ );
	
	PetscFunctionReturn(0);
}

/*45*/
PetscErrorCode MatGetRowMax_Schur(Mat mat,Vec v,PetscInt idx[])
{
	PetscFunctionBegin; SETERRQ1( PETSC_ERR_SUP,"Operation: %s is not implemented", __func__ );
	
	PetscFunctionReturn(0);
}
PetscErrorCode MatScale_Schur(Mat mat,PetscScalar a)
{
	PetscFunctionBegin;
	
	MatSchurSetScalar(mat,a);
	
	PetscFunctionReturn(0);
}
PetscErrorCode MatShift_Schur( Mat mat, PetscScalar a )
{
	PetscFunctionBegin; SETERRQ1( PETSC_ERR_SUP,"Operation: %s is not implemented", __func__ );
	
	PetscFunctionReturn(0);
}

PetscErrorCode MatDiagonalSet_Schur( Mat Y, Vec D, InsertMode is )
{
	PetscFunctionBegin; SETERRQ1( PETSC_ERR_SUP,"Operation: %s is not implemented", __func__ );
	
	PetscFunctionReturn(0);
}
PetscErrorCode MatILUDTFactor_Schur(Mat mat,IS row,IS col,MatFactorInfo *info,Mat *fact)
{
	PetscFunctionBegin; SETERRQ1( PETSC_ERR_SUP,"Operation: %s is not implemented", __func__ );
	
	PetscFunctionReturn(0);
}

/*50*/
PetscErrorCode MatSetBlockSize_Schur(Mat mat,PetscInt bs)
{
	PetscFunctionBegin; SETERRQ1( PETSC_ERR_SUP,"Operation: %s is not implemented", __func__ );
	
	PetscFunctionReturn(0);
}
PetscErrorCode MatGetRowIJ_Schur(Mat mat,PetscInt shift,PetscTruth symmetric,PetscTruth blockcompressed,PetscInt *n,PetscInt *ia[],PetscInt* ja[],PetscTruth *done)
{
	PetscFunctionBegin; SETERRQ1( PETSC_ERR_SUP,"Operation: %s is not implemented", __func__ );
	
	PetscFunctionReturn(0);
}
PetscErrorCode MatRestorRowIJ_Schur(Mat mat,PetscInt shift,PetscTruth symmetric,PetscTruth blockcompressed,PetscInt *n,PetscInt *ia[],PetscInt* ja[],PetscTruth *done)
{
	PetscFunctionBegin; SETERRQ1( PETSC_ERR_SUP,"Operation: %s is not implemented", __func__ );
	
	PetscFunctionReturn(0);
}
PetscErrorCode MatGetColumnIJ_Schur(Mat mat,PetscInt shift,PetscTruth symmetric,PetscTruth blockcompressed,PetscInt *n,PetscInt *ia[],PetscInt* ja[],PetscTruth *done)
{
	PetscFunctionBegin; SETERRQ1( PETSC_ERR_SUP,"Operation: %s is not implemented", __func__ );
	
	PetscFunctionReturn(0);
}
PetscErrorCode MatRestoreColumnIJ_Schur(Mat mat,PetscInt shift,PetscTruth symmetric,PetscTruth blockcompressed,PetscInt *n,PetscInt *ia[],PetscInt* ja[],PetscTruth *done)
{
	PetscFunctionBegin; SETERRQ1( PETSC_ERR_SUP,"Operation: %s is not implemented", __func__ );
	
	PetscFunctionReturn(0);
}

/*55*/

PetscErrorCode MatFDColoringCreate_Schur(Mat mat,ISColoring iscoloring,MatFDColoring color)
{
	PetscFunctionBegin; SETERRQ1( PETSC_ERR_SUP,"Operation: %s is not implemented", __func__ );
	
	PetscFunctionReturn(0);
}

PetscErrorCode MatColoringPatch_Schur(Mat mat,PetscInt ncolors,PetscInt n,ISColoringValue colorarray[],ISColoring *iscoloring)
{
	PetscFunctionBegin; SETERRQ1( PETSC_ERR_SUP,"Operation: %s is not implemented", __func__ );
	
	PetscFunctionReturn(0);
}
PetscErrorCode MatSetUnfactored_Schur(Mat mat)
{
	PetscFunctionBegin; SETERRQ1( PETSC_ERR_SUP,"Operation: %s is not implemented", __func__ );
	
	PetscFunctionReturn(0);
}


PetscErrorCode MatPermute_Schur( Mat mat, IS row, IS col, Mat *B )
{
	PetscFunctionBegin; SETERRQ1( PETSC_ERR_SUP,"Operation: %s is not implemented", __func__ );
	
	PetscFunctionReturn(0);
}
PetscErrorCode MatSetValuesBlocked_Schur(Mat mat,PetscInt m,const PetscInt idxm[],PetscInt n,const PetscInt idxn[],const PetscScalar v[],InsertMode addv)
{
	PetscFunctionBegin; SETERRQ1( PETSC_ERR_SUP,"Operation: %s is not implemented", __func__ );
	
	PetscFunctionReturn(0);
}

/*60*/
PetscErrorCode MatGetSubMatrix_Schur(Mat mat,IS isrow,IS iscol,PetscInt csize,MatReuse cll,Mat *newmat)
{
	PetscFunctionBegin; SETERRQ1( PETSC_ERR_SUP,"Operation: %s is not implemented", __func__ );
	
	PetscFunctionReturn(0);
}
/*
PetscErrorCode MatDestroy_Schur( Mat mat )
{
	PetscFunctionBegin; SETERRQ1( PETSC_ERR_SUP,"Operation: %s is not implemented", __func__ );
	
	PetscFunctionReturn(0);
}
*/
PetscErrorCode MatView_Schur( Mat mat, PetscViewer viewer )
{
	Mat_Schur s = (Mat_Schur)mat->data;
	PetscTruth isascii;
	PetscViewerFormat format;
	
	PetscTypeCompare((PetscObject)viewer,PETSC_VIEWER_ASCII,&isascii);
	if (isascii) {
		
		PetscViewerGetFormat(viewer,&format);
		PetscViewerSetFormat(viewer,PETSC_VIEWER_ASCII_INFO);
		
		PetscViewerASCIIPrintf(viewer,"complement type: %s\n", s->stype );
		if (s->scale_set) {
			PetscViewerASCIIPrintf(viewer,"scaling factor=%lf\n", s->alpha );
		}
		
		if (s->A11) {
			PetscViewerASCIIPrintf(viewer,"A11:---------------------\n");
			MatView(s->A11,viewer);
		}
		if (s->A12) {
			PetscViewerASCIIPrintf(viewer,"A12:---------------------\n");
			MatView(s->A12,viewer);
		}
		if (s->A21) {
			PetscViewerASCIIPrintf(viewer,"A21:---------------------\n");
			MatView(s->A21,viewer);
		}
		if (s->A22) {
			PetscViewerASCIIPrintf(viewer,"A22:---------------------\n");
			MatView(s->A22,viewer);
		}
		
		if (s->ksp) {
			PetscViewerASCIIPrintf(viewer,"Inner KSP:---------------------\n");
			KSPView(s->ksp,viewer);
		}
		
		PetscViewerSetFormat(viewer,format);
	}
	
	PetscFunctionReturn(0);
}
PetscErrorCode MatConvertFrom_Schur(Mat mat, MatType newtype,MatReuse reuse,Mat *M)
{
	PetscFunctionBegin; SETERRQ1( PETSC_ERR_SUP,"Operation: %s is not implemented", __func__ );
	
	PetscFunctionReturn(0);
}
PetscErrorCode MatUseScaledForm_Schur( Mat mat, PetscTruth scaled )
{
	PetscFunctionBegin; SETERRQ1( PETSC_ERR_SUP,"Operation: %s is not implemented", __func__ );
	
	PetscFunctionReturn(0);
}

/*65*/
PetscErrorCode MatScaleSystem_Schur( Mat mat, Vec b, Vec x )
{
	PetscFunctionBegin; SETERRQ1( PETSC_ERR_SUP,"Operation: %s is not implemented", __func__ );
	
	PetscFunctionReturn(0);
}
PetscErrorCode MatUnScaleSystem_Schur(Mat mat,Vec b,Vec x)
{
	PetscFunctionBegin; SETERRQ1( PETSC_ERR_SUP,"Operation: %s is not implemented", __func__ );
	
	PetscFunctionReturn(0);
}
PetscErrorCode MatSetLocalToGlobalMapping_Schur(Mat x,ISLocalToGlobalMapping mapping)
{
	PetscFunctionBegin; SETERRQ1( PETSC_ERR_SUP,"Operation: %s is not implemented", __func__ );
	
	PetscFunctionReturn(0);
}
PetscErrorCode MatSetValuesLocal_Schur(Mat mat,PetscInt nrow,const PetscInt irow[],PetscInt ncol,const PetscInt icol[],const PetscScalar y[],InsertMode addv)
{
	PetscFunctionBegin; SETERRQ1( PETSC_ERR_SUP,"Operation: %s is not implemented", __func__ );
	
	PetscFunctionReturn(0);
}
PetscErrorCode MatZeroRowsLocal_Schur(Mat mat,PetscInt numRows,const PetscInt rows[],PetscScalar diag)
{
	PetscFunctionBegin; SETERRQ1( PETSC_ERR_SUP,"Operation: %s is not implemented", __func__ );
	
	PetscFunctionReturn(0);
}

/*70*/
PetscErrorCode MatGetRowMaxAbs_Schur(Mat mat,Vec v,PetscInt idx[])
{
	PetscFunctionBegin; SETERRQ1( PETSC_ERR_SUP,"Operation: %s is not implemented", __func__ );
	
	PetscFunctionReturn(0);
}
PetscErrorCode MatConvert_Schur(Mat mat, MatType newtype,MatReuse reuse,Mat *M)
{
	PetscFunctionBegin; SETERRQ1( PETSC_ERR_SUP,"Operation: %s is not implemented", __func__ );
	
	PetscFunctionReturn(0);
}
PetscErrorCode MatSetColoring_Schur( Mat mat, ISColoring colour )
{
	PetscFunctionBegin; SETERRQ1( PETSC_ERR_SUP,"Operation: %s is not implemented", __func__ );
	
	PetscFunctionReturn(0);
}
PetscErrorCode MatSetValuesAdic_Schur(Mat mat,void *v)
{
	PetscFunctionBegin; SETERRQ1( PETSC_ERR_SUP,"Operation: %s is not implemented", __func__ );
	
	PetscFunctionReturn(0);
}
PetscErrorCode MatSetValuesAdifor_Schur(Mat mat,PetscInt nl,void *v)
{
	PetscFunctionBegin; SETERRQ1( PETSC_ERR_SUP,"Operation: %s is not implemented", __func__ );
	
	PetscFunctionReturn(0);
}


/*75*/
PetscErrorCode MatFDColoringApply_Schur(Mat J,MatFDColoring coloring,Vec x1,MatStructure *flag,void *sctx)
{
	PetscFunctionBegin; SETERRQ1( PETSC_ERR_SUP,"Operation: %s is not implemented", __func__ );
	
	PetscFunctionReturn(0);
}
PetscErrorCode MatSetFromOptions_Schur( Mat mat )
{
	PetscTruth     flg;
	PetscScalar    val;
	char           type[PETSC_MAX_PATH_LEN];
    PetscErrorCode ierr;
	
	PetscFunctionBegin; 
	ierr = PetscOptionsBegin(PETSC_COMM_WORLD, PETSC_NULL, "Schur Options", "PC");CHKERRQ(ierr);
	PetscOptionsScalar("-mat_schur_scalar","Scaling factor","MatSchurSetScalar",1.0,&val,&flg);
	if (flg) {			MatSchurSetScalar(mat,val);					}
	PetscOptionsString("-mat_schur_complement_type","Schur complemenet","MatSchurSetSchurComplementType",_MatSchurComplementType[0],type,PETSC_MAX_PATH_LEN-1,&flg);
	if (flg) {			MatSchurSetSchurComplementType(mat,type);								}
	ierr = PetscOptionsEnd();CHKERRQ(ierr);
	PetscFunctionReturn(0);
}
PetscErrorCode MatMultConstrained_Schur(Mat mat,Vec x,Vec y)
{
	PetscFunctionBegin; SETERRQ1( PETSC_ERR_SUP,"Operation: %s is not implemented", __func__ );
	
	PetscFunctionReturn(0);
}
PetscErrorCode MatMultTransposeConstrained_Schur(Mat mat,Vec x,Vec y)
{
	PetscFunctionBegin; SETERRQ1( PETSC_ERR_SUP,"Operation: %s is not implemented", __func__ );
	
	PetscFunctionReturn(0);
}
PetscErrorCode MatILUFactorSymbolicConstrained_Schur(Mat mat,IS row,IS col,double d,PetscInt i,PetscInt j,Mat *fact)
{
	PetscFunctionBegin; SETERRQ1( PETSC_ERR_SUP,"Operation: %s is not implemented", __func__ );
	
	PetscFunctionReturn(0);
}


/*80*/
PetscErrorCode MatPermuteSparsify_Schur(Mat A, PetscInt band, PetscReal frac, PetscReal tol, IS rowp, IS colp, Mat *B)
{
	PetscFunctionBegin; SETERRQ1( PETSC_ERR_SUP,"Operation: %s is not implemented", __func__ );
	
	PetscFunctionReturn(0);
}
PetscErrorCode MatMults_Schur(Mat A,Vecs x,Vecs y)
{
	PetscFunctionBegin; SETERRQ1( PETSC_ERR_SUP,"Operation: %s is not implemented", __func__ );
	
	PetscFunctionReturn(0);
}
PetscErrorCode MatSolves_Schur(Mat A,Vecs b,Vecs x)
{
	PetscFunctionBegin; SETERRQ1( PETSC_ERR_SUP,"Operation: %s is not implemented", __func__ );
	
	PetscFunctionReturn(0);
}
PetscErrorCode MatGetInertia_Schur(Mat mat,PetscInt *nneg,PetscInt *nzero,PetscInt *npos)
{
	PetscFunctionBegin; SETERRQ1( PETSC_ERR_SUP,"Operation: %s is not implemented", __func__ );
	
	PetscFunctionReturn(0);
}
PetscErrorCode MatLoad_Schur(PetscViewer viewer, MatType outtype,Mat *newmat)
{
	PetscFunctionBegin; SETERRQ1( PETSC_ERR_SUP,"Operation: %s is not implemented", __func__ );
	
	PetscFunctionReturn(0);
}


/*85*/
PetscErrorCode MatIsSymmetric_Schur(Mat A,PetscReal tol,PetscTruth *flg)
{
	PetscFunctionBegin; SETERRQ1( PETSC_ERR_SUP,"Operation: %s is not implemented", __func__ );
	
	PetscFunctionReturn(0);
}
PetscErrorCode MatIsHermitian_Schur(Mat A,PetscTruth *flg)
{
	PetscFunctionBegin; SETERRQ1( PETSC_ERR_SUP,"Operation: %s is not implemented", __func__ );
	
	PetscFunctionReturn(0);
}
PetscErrorCode MatIsStructurallySymmetric_Schur(Mat A,PetscTruth *flg)
{
	PetscFunctionBegin; SETERRQ1( PETSC_ERR_SUP,"Operation: %s is not implemented", __func__ );
	
	PetscFunctionReturn(0);
}
PetscErrorCode MatPBRelax_Schur(Mat mat,Vec b,PetscReal omega,MatSORType flag,PetscReal shift,PetscInt its,PetscInt lits,Vec x)
{
	PetscFunctionBegin; SETERRQ1( PETSC_ERR_SUP,"Operation: %s is not implemented", __func__ );
	
	PetscFunctionReturn(0);
}
PetscErrorCode MatGetVecs_Schur(Mat mat,Vec *right,Vec *left)
{
	Mat_Schur s = (Mat_Schur)mat->data;
	PetscTruth flg;
	
	
	PetscFunctionBegin;
	
	PetscStrcmp( s->stype, "MatSchur_A11", &flg );
	if( flg == PETSC_TRUE ) {
		if( left != PETSC_NULL ) {		MatGetVecs( s->A21, PETSC_NULL, left );		}
		if( right != PETSC_NULL ){		MatGetVecs( s->A21, PETSC_NULL, right );	}
	}
	
	PetscStrcmp( s->stype, "MatSchur_A22", &flg );
	if( flg == PETSC_TRUE ) {
		
		if( left != PETSC_NULL ) {		MatGetVecs( s->A12, PETSC_NULL, left );		}
		if( right != PETSC_NULL ){		MatGetVecs( s->A12, PETSC_NULL, right );	}
		
	}
	
	PetscFunctionReturn(0);
}



/*90*/
PetscErrorCode MatMatMult_Schur(Mat A,Mat B,MatReuse scall,PetscReal fill,Mat *C) 
{
	PetscFunctionBegin; SETERRQ1( PETSC_ERR_SUP,"Operation: %s is not implemented", __func__ );
	
	PetscFunctionReturn(0);
}
PetscErrorCode MatMatMultSymbolic_Schur(Mat A,Mat B,PetscReal fill,Mat *C) 
{
	PetscFunctionBegin; SETERRQ1( PETSC_ERR_SUP,"Operation: %s is not implemented", __func__ );
	
	PetscFunctionReturn(0);
}
PetscErrorCode MatMatMultNumeric_Schur(Mat A,Mat B,Mat C)
{
	PetscFunctionBegin; SETERRQ1( PETSC_ERR_SUP,"Operation: %s is not implemented", __func__ );
	
	PetscFunctionReturn(0);
}
PetscErrorCode MatPtAP_Schur(Mat A,Mat P,MatReuse scall,PetscReal fill,Mat *C) 
{
	PetscFunctionBegin; SETERRQ1( PETSC_ERR_SUP,"Operation: %s is not implemented", __func__ );
	
	PetscFunctionReturn(0);
}
PetscErrorCode MatPtAPSymbolic_Schur(Mat A,Mat P,PetscReal fill,Mat *C) 
{
	PetscFunctionBegin; SETERRQ1( PETSC_ERR_SUP,"Operation: %s is not implemented", __func__ );
	
	PetscFunctionReturn(0);
}


/*95*/
PetscErrorCode MatPtAPNumeric_Schur(Mat A,Mat P,Mat C) 
{
	PetscFunctionBegin; SETERRQ1( PETSC_ERR_SUP,"Operation: %s is not implemented", __func__ );
	
	PetscFunctionReturn(0);
}
PetscErrorCode MatMatMultTranspose_Schur(Mat A,Mat B,MatReuse scall,PetscReal fill,Mat *C) 
{
	PetscFunctionBegin; SETERRQ1( PETSC_ERR_SUP,"Operation: %s is not implemented", __func__ );
	
	PetscFunctionReturn(0);
}
PetscErrorCode MatMatMultTransposeSymbolic_Schur(Mat A,Mat B,PetscReal fill,Mat *C) 
{
	PetscFunctionBegin; SETERRQ1( PETSC_ERR_SUP,"Operation: %s is not implemented", __func__ );
	
	PetscFunctionReturn(0);
}
PetscErrorCode MatMatMultTransposeNumeric_Schur(Mat A,Mat B,Mat C)
{
	PetscFunctionBegin; SETERRQ1( PETSC_ERR_SUP,"Operation: %s is not implemented", __func__ );
	
	PetscFunctionReturn(0);
}
PetscErrorCode MatPtAPSymbolic__SEQAIJ_Schur(Mat A,Mat P,PetscReal fill,Mat *C) 
{
	PetscFunctionBegin; SETERRQ1( PETSC_ERR_SUP,"Operation: %s is not implemented", __func__ );
	
	PetscFunctionReturn(0);
}


/*100*/
PetscErrorCode MatPtAPNumeric_SEQAIJ_Schur(Mat A,Mat P,Mat C) 
{
	PetscFunctionBegin; SETERRQ1( PETSC_ERR_SUP,"Operation: %s is not implemented", __func__ );
	
	PetscFunctionReturn(0);
}
PetscErrorCode MatPtAPSymbolic_MPIAIJ_Schur(Mat A,Mat P,PetscReal fill,Mat *C) 
{
	PetscFunctionBegin; SETERRQ1( PETSC_ERR_SUP,"Operation: %s is not implemented", __func__ );
	
	PetscFunctionReturn(0);
}
PetscErrorCode MatPtAPNumeric_MPIAIJ_Schur(Mat A,Mat P,Mat C) 
{
	PetscFunctionBegin; SETERRQ1( PETSC_ERR_SUP,"Operation: %s is not implemented", __func__ );
	
	PetscFunctionReturn(0);
}
PetscErrorCode MatConjugate_Schur(Mat mat)
{
	PetscFunctionBegin; SETERRQ1( PETSC_ERR_SUP,"Operation: %s is not implemented", __func__ );
	
	PetscFunctionReturn(0);
}
PetscErrorCode MatSetSizes_Schur(Mat mat,PetscInt m,PetscInt n,PetscInt M,PetscInt N)
{
	PetscFunctionBegin; SETERRQ1( PETSC_ERR_SUP,"Operation: %s is not implemented", __func__ );
	
	PetscFunctionReturn(0);
}


/*105*/
PetscErrorCode MatSetValuesRow_Schur(Mat mat,PetscInt row,const PetscScalar v[])
{
	PetscFunctionBegin; SETERRQ1( PETSC_ERR_SUP,"Operation: %s is not implemented", __func__ );
	
	PetscFunctionReturn(0);
}
PetscErrorCode MatRealPart_Schur(Mat mat)
{
	PetscFunctionBegin; SETERRQ1( PETSC_ERR_SUP,"Operation: %s is not implemented", __func__ );
	
	PetscFunctionReturn(0);
}
PetscErrorCode MatImaginaryPart_Schur(Mat mat)
{
	PetscFunctionBegin; SETERRQ1( PETSC_ERR_SUP,"Operation: %s is not implemented", __func__ );
	
	PetscFunctionReturn(0);
}

PetscErrorCode MatGetRowUpperTriangular_Schur(Mat mat)
{
	PetscFunctionBegin; SETERRQ1( PETSC_ERR_SUP,"Operation: %s is not implemented", __func__ );
	
	PetscFunctionReturn(0);
}

PetscErrorCode MatRestoreRowUpperTriangular_Schur(Mat mat)
{
	PetscFunctionBegin; SETERRQ1( PETSC_ERR_SUP,"Operation: %s is not implemented", __func__ );
	
	PetscFunctionReturn(0);
}


/*110*/
PetscErrorCode MatMatSolve_Schur(Mat mat,Mat b,Mat x)
{
	PetscFunctionBegin; SETERRQ1( PETSC_ERR_SUP,"Operation: %s is not implemented", __func__ );
	
	PetscFunctionReturn(0);
}
PetscErrorCode MatGetRedundantMatrix_Schur(Mat mat,PetscInt nsubcomm,MPI_Comm subcomm,PetscInt mlocal_red,MatReuse reuse,Mat *matredundant)
{
	PetscFunctionBegin; SETERRQ1( PETSC_ERR_SUP,"Operation: %s is not implemented", __func__ );
	
	PetscFunctionReturn(0);
}
PetscErrorCode MatGetRowMin_Schur(Mat mat,Vec v,PetscInt idx[])
{
	PetscFunctionBegin; SETERRQ1( PETSC_ERR_SUP,"Operation: %s is not implemented", __func__ );
	
	PetscFunctionReturn(0);
}
PetscErrorCode MatGetColumnVector_Schur(Mat A,Vec yy,PetscInt col)
{
	PetscFunctionBegin; SETERRQ1( PETSC_ERR_SUP,"Operation: %s is not implemented", __func__ );
	
	PetscFunctionReturn(0);
}

